@@ -423,7 +423,7 @@ std::string CWriter::MangleName(std::string_view name) {
   std::string result = "Z_";
 
   if (!name.empty()) {
-    for (char c : name) {
+    for (unsigned char c : name) {
       if ((isalnum(c) && c != kPrefix) || c == '_') {
         result += c;
       } else {
@@ -461,9 +461,9 @@ std::string CWriter::LegalizeName(std::string_view name) {
     return "_";
 
   std::string result;
-  result = isalpha(name[0]) ? name[0] : '_';
+  result = isalpha(static_cast<unsigned char>(name[0])) ? name[0] : '_';
   for (size_t i = 1; i < name.size(); ++i)
-    result += isalnum(name[i]) ? name[i] : '_';
+    result += isalnum(static_cast<unsigned char>(name[i])) ? name[i] : '_';
 
   // In addition to containing valid characters for C, we must also avoid
   // colliding with things C cares about, such as reserved words (e.g. "void")
@@ -833,28 +833,33 @@ void CWriter::WriteMultivalueTypes() {
     // incidentally they also mean we don't have to bother with deduplication
     Write("#ifndef ", name, Newline());
     Write("#define ", name, " ", name, Newline());
-    Write("struct ", name, " {", Newline());
+    Write("struct ", name, " ", OpenBrace());
     for (Index i = 0; i < num_results; ++i) {
       Type type = func_type->GetResultType(i);
-      Write("  ", type);
+      Write(type);
       Writef(" %c%d;", MangleType(type), i);
       Write(Newline());
     }
-    Write("};", Newline(), "#endif  /* ", name, " */", Newline());
+    Write(CloseBrace(), ";", Newline(), "#endif  /* ", name, " */", Newline());
   }
 }
 
 void CWriter::WriteFuncTypes() {
-  Write(Newline());
-  Writef("static u32 func_types[%" PRIzd "];", module_->types.size());
-  Write(Newline(), Newline());
-  Write("static void init_func_types(void) {", Newline());
+  if (module_->types.size()) {
+    Writef("static u32 func_types[%" PRIzd "];", module_->types.size());
+    Write(Newline());
+  }
+  Write("static void init_func_types(void) ", OpenBrace());
+  if (!module_->types.size()) {
+    Write(CloseBrace(), Newline());
+    return;
+  }
   Index func_type_index = 0;
   for (TypeEntry* type : module_->types) {
     FuncType* func_type = cast<FuncType>(type);
     Index num_params = func_type->GetNumParams();
     Index num_results = func_type->GetNumResults();
-    Write("  func_types[", func_type_index, "] = wasm_rt_register_func_type(",
+    Write("func_types[", func_type_index, "] = wasm_rt_register_func_type(",
           num_params, ", ", num_results);
     for (Index i = 0; i < num_params; ++i) {
       Write(", ", TypeEnum(func_type->GetParamType(i)));
@@ -867,7 +872,7 @@ void CWriter::WriteFuncTypes() {
     Write(");", Newline());
     ++func_type_index;
   }
-  Write("}", Newline());
+  Write(CloseBrace(), Newline());
 }
 
 void CWriter::WriteImports() {
@@ -1017,8 +1022,9 @@ void CWriter::WriteMemory(const std::string& name) {
 }
 
 void CWriter::WriteTables() {
-  if (module_->tables.size() == module_->num_table_imports)
+  if (module_->tables.size() == module_->num_table_imports) {
     return;
+  }
 
   Write(Newline());
 
@@ -1048,17 +1054,22 @@ void CWriter::WriteDataInitializers() {
       Write(Newline());
     } else {
       for (const DataSegment* data_segment : module_->data_segments) {
-        Write(Newline(), "static const u8 data_segment_data_",
-              data_segment_index, "[] = ", OpenBrace());
-        size_t i = 0;
-        for (uint8_t x : data_segment->data) {
-          Writef("0x%02x, ", x);
-          if ((++i % 12) == 0)
+        if (!data_segment->data.size()) {
+          Write(Newline(), "static const u8* data_segment_data_",
+                data_segment_index, " = NULL;", Newline());
+        } else {
+          Write(Newline(), "static const u8 data_segment_data_",
+                data_segment_index, "[] = ", OpenBrace());
+          size_t i = 0;
+          for (uint8_t x : data_segment->data) {
+            Writef("0x%02x, ", x);
+            if ((++i % 12) == 0)
+              Write(Newline());
+          }
+          if (i > 0)
             Write(Newline());
+          Write(CloseBrace(), ";", Newline());
         }
-        if (i > 0)
-          Write(Newline());
-        Write(CloseBrace(), ";", Newline());
         ++data_segment_index;
       }
     }
@@ -1082,7 +1093,8 @@ void CWriter::WriteDataInitializers() {
     Write("LOAD_DATA(", ExternalRef(memory->name), ", ");
     WriteInitExpr(data_segment->offset);
     Write(", data_segment_data_", data_segment_index, ", ",
-          data_segment->data.size(), ");", Newline());
+          data_segment->data.size());
+    Write(");", Newline());
     ++data_segment_index;
   }
 
@@ -1090,9 +1102,18 @@ void CWriter::WriteDataInitializers() {
 }
 
 void CWriter::WriteElemInitializers() {
+  Write(Newline(), "static void init_table(void) ", OpenBrace());
+
+  if (!module_->types.size()) {
+    // If there are no types there cannot be any table entries either.
+    for (const ElemSegment* elem_segment : module_->elem_segments) {
+      assert(elem_segment->elem_exprs.size() == 0);
+    }
+    Write(CloseBrace(), Newline());
+    return;
+  }
   const Table* table = module_->tables.empty() ? nullptr : module_->tables[0];
 
-  Write(Newline(), "static void init_table(void) ", OpenBrace());
   Write("uint32_t offset;", Newline());
   if (table && module_->num_table_imports == 0) {
     uint32_t max =
@@ -1498,6 +1519,9 @@ void CWriter::Write(const ExprList& exprs) {
         break;
       }
 
+      case ExprType::CodeMetadata:
+        break;
+
       case ExprType::Compare:
         Write(*cast<CompareExpr>(&expr));
         break;
@@ -2190,11 +2214,11 @@ void CWriter::Write(const UnaryExpr& expr) {
       break;
 
     case Opcode::F32Abs:
-      WriteSimpleUnaryExpr(expr.opcode, "fabsf");
+      WriteSimpleUnaryExpr(expr.opcode, "wasm_rt_fabsf");
       break;
 
     case Opcode::F64Abs:
-      WriteSimpleUnaryExpr(expr.opcode, "fabs");
+      WriteSimpleUnaryExpr(expr.opcode, "wasm_rt_fabs");
       break;
 
     case Opcode::F32Sqrt:
@@ -2222,19 +2246,19 @@ void CWriter::Write(const UnaryExpr& expr) {
       break;
 
     case Opcode::F32Trunc:
-      WriteSimpleUnaryExpr(expr.opcode, "truncf");
+      WriteSimpleUnaryExpr(expr.opcode, "wasm_rt_truncf");
       break;
 
     case Opcode::F64Trunc:
-      WriteSimpleUnaryExpr(expr.opcode, "trunc");
+      WriteSimpleUnaryExpr(expr.opcode, "wasm_rt_trunc");
       break;
 
     case Opcode::F32Nearest:
-      WriteSimpleUnaryExpr(expr.opcode, "nearbyintf");
+      WriteSimpleUnaryExpr(expr.opcode, "wasm_rt_nearbyintf");
       break;
 
     case Opcode::F64Nearest:
-      WriteSimpleUnaryExpr(expr.opcode, "nearbyint");
+      WriteSimpleUnaryExpr(expr.opcode, "wasm_rt_nearbyint");
       break;
 
     case Opcode::I32Extend8S:
