@@ -19,9 +19,11 @@ import argparse
 import io
 import json
 import os
+import platform
 import re
 import struct
 import sys
+import shlex
 
 import find_exe
 import utils
@@ -29,6 +31,8 @@ from utils import Error
 
 SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
 WASM2C_DIR = os.path.join(find_exe.REPO_ROOT_DIR, 'wasm2c')
+IS_WINDOWS = sys.platform == 'win32'
+IS_MACOS = platform.mac_ver()[0] != ''
 
 
 def ReinterpretF32(f32_bits):
@@ -77,7 +81,7 @@ def F64ToC(f64_bits):
     elif f64_bits == F64_SIGN_BIT:
         return '-0.0'
     else:
-        return '%.17g' % ReinterpretF64(f64_bits)
+        return '%#.17gL' % ReinterpretF64(f64_bits)
 
 
 def MangleType(t):
@@ -334,17 +338,45 @@ class CWriter(object):
 
 
 def Compile(cc, c_filename, out_dir, *args):
-    o_filename = utils.ChangeDir(utils.ChangeExt(c_filename, '.o'), out_dir)
-    cc.RunWithArgs('-c', c_filename, '-o', o_filename, *args)
+    if IS_WINDOWS:
+        ext = '.obj'
+    else:
+        ext = '.o'
+    o_filename = utils.ChangeDir(utils.ChangeExt(c_filename, ext), out_dir)
+    args = list(args)
+    if IS_WINDOWS:
+        args += ['/nologo', '/MDd', '/c', c_filename, '/Fo' + o_filename]
+    else:
+        args += ['-c', c_filename, '-o', o_filename,
+                 '-Wall', '-Werror', '-Wno-unused',
+                 '-Wno-tautological-constant-out-of-range-compare',
+                 '-std=c99', '-D_DEFAULT_SOURCE']
+    # Use RunWithArgsForStdout and discard stdout because cl.exe
+    # unconditionally prints the name of input files on stdout
+    # and we don't want that to be part of our stdout.
+    cc.RunWithArgsForStdout(*args)
     return o_filename
 
 
-def Link(cc, o_filenames, main_exe, *args):
-    args = ['-o', main_exe] + o_filenames + list(args)
-    cc.RunWithArgs(*args)
+def Link(cc, o_filenames, main_exe, *extra_args):
+    args = o_filenames
+    if IS_WINDOWS:
+        # Windows default to 1Mb of stack but `spec/skip-stack-guard-page.wast`
+        # uses more than this.  Set to 8Mb for parity with linux.
+        args += ['/nologo', '/MDd', '/link', '/stack:8388608', '/out:' + main_exe]
+    else:
+        args += ['-o', main_exe]
+    args += list(extra_args)
+    # Use RunWithArgsForStdout and discard stdout because cl.exe
+    # unconditionally prints the name of input files on stdout
+    # and we don't want that to be part of our stdout.
+    cc.RunWithArgsForStdout(*args)
 
 
 def main(args):
+    default_compiler = 'cc'
+    if IS_WINDOWS:
+        default_compiler = 'cl.exe'
     parser = argparse.ArgumentParser()
     parser.add_argument('-o', '--out-dir', metavar='PATH',
                         help='output directory for files.')
@@ -356,7 +388,8 @@ def main(args):
     parser.add_argument('--wasmrt-dir', metavar='PATH',
                         help='directory with wasm-rt files', default=WASM2C_DIR)
     parser.add_argument('--cc', metavar='PATH',
-                        help='the path to the C compiler', default='cc')
+                        help='the path to the C compiler',
+                        default=default_compiler)
     parser.add_argument('--cflags', metavar='FLAGS',
                         help='additional flags for C compiler.',
                         action='append', default=[])
@@ -377,6 +410,9 @@ def main(args):
                         help='print the commands that are run.',
                         action='store_true')
     parser.add_argument('file', help='wast file.')
+    parser.add_argument('--enable-multi-memory', action='store_true')
+    parser.add_argument('--disable-bulk-memory', action='store_true')
+    parser.add_argument('--disable-reference-types', action='store_true')
     options = parser.parse_args(args)
 
     with utils.TempDirectory(options.out_dir, 'run-spec-wasm2c-') as out_dir:
@@ -385,7 +421,11 @@ def main(args):
             find_exe.GetWast2JsonExecutable(options.bindir),
             error_cmdline=options.error_cmdline)
         wast2json.verbose = options.print_cmd
-        wast2json.AppendOptionalArgs({'-v': options.verbose})
+        wast2json.AppendOptionalArgs({
+            '-v': options.verbose,
+            '--enable-multi-memory': options.enable_multi_memory,
+            '--disable-bulk-memory': options.disable_bulk_memory,
+            '--disable-reference-types': options.disable_reference_types})
 
         json_file_path = utils.ChangeDir(
             utils.ChangeExt(options.file, '.json'), out_dir)
@@ -395,11 +435,15 @@ def main(args):
             find_exe.GetWasm2CExecutable(options.bindir),
             error_cmdline=options.error_cmdline)
         wasm2c.verbose = options.print_cmd
+        wasm2c.AppendOptionalArgs({
+            '--enable-multi-memory': options.enable_multi_memory})
 
-        cc = utils.Executable(options.cc, *options.cflags)
+        options.cflags += shlex.split(os.environ.get('WASM2C_CFLAGS', ''))
+        cc = utils.Executable(options.cc, *options.cflags, forward_stderr=True,
+                              forward_stdout=False)
         cc.verbose = options.print_cmd
 
-        with open(json_file_path) as json_file:
+        with open(json_file_path, encoding='utf-8') as json_file:
             spec_json = json.load(json_file)
 
         prefix = ''
@@ -418,10 +462,6 @@ def main(args):
         o_filenames = []
         includes = '-I%s' % options.wasmrt_dir
 
-        # Compile wasm-rt-impl.
-        wasm_rt_impl_c = os.path.join(options.wasmrt_dir, 'wasm-rt-impl.c')
-        o_filenames.append(Compile(cc, wasm_rt_impl_c, out_dir, includes))
-
         for i, wasm_filename in enumerate(cwriter.GetModuleFilenames()):
             wasm_filename = os.path.join(out_dir, wasm_filename)
             c_filename = utils.ChangeExt(wasm_filename, '.c')
@@ -431,12 +471,24 @@ def main(args):
                 o_filenames.append(Compile(cc, c_filename, out_dir, includes, defines))
 
         if options.compile:
-            o_filenames.append(Compile(cc, main_filename, out_dir, includes, defines))
-            main_exe = utils.ChangeExt(json_file_path, '')
-            Link(cc, o_filenames, main_exe, '-lm')
+            # Compile wasm-rt-impl.
+            wasm_rt_impl_c = os.path.join(options.wasmrt_dir, 'wasm-rt-impl.c')
+            o_filenames.append(Compile(cc, wasm_rt_impl_c, out_dir, includes))
 
-        if options.compile and options.run:
-            utils.Executable(main_exe, forward_stdout=True).RunWithArgs()
+            # Compile and link -main test run entry point
+            o_filenames.append(Compile(cc, main_filename, out_dir, includes))
+            if IS_WINDOWS:
+                exe_ext = '.exe'
+                libs = []
+            else:
+                exe_ext = ''
+                libs = ['-lm']
+            main_exe = utils.ChangeExt(json_file_path, exe_ext)
+            Link(cc, o_filenames, main_exe, *libs)
+
+            # Run the resulting binary
+            if options.run:
+                utils.Executable(main_exe, forward_stdout=True).RunWithArgs()
 
     return 0
 
